/*
 * piTclCvtUtil.c --
 *
 *	Implementation of conversion utilities.
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: cvtUtil.c,v 1.1 1999/03/27 13:26:31 aku Exp $
 */

#include "internal.h"

/*
 *------------------------------------------------------*
 * Definition for conversion of record attributes.
 */

static TclUtil_Bit recABits [] = {
  {"archive", dlpRecAttrArchived},
  {"delete",  dlpRecAttrDeleted},
  {"dirty",   dlpRecAttrDirty},
  {"locked",  dlpRecAttrBusy},
  {"secret",  dlpRecAttrSecret},
  {NULL,      0}
};

static TclUtil_Bitset recordAttributes = {
  "record attribute",
  NULL,
  recABits
};

/*
 *------------------------------------------------------*
 * Definition for conversion of possible open flags for
 * databases
 */

static TclUtil_Bit oFBits [] = {
  {"exclusive", dlpOpenExclusive},
  {"read",      dlpOpenRead},
  {"secret",    dlpOpenSecret},
  {"write",     dlpOpenWrite},
  {NULL,      0}
};

static TclUtil_Bitset openFlags = {
  "open mode",
  NULL,
  oFBits
};

/*
 *------------------------------------------------------*
 * Definition for conversion of database flags
 */

static TclUtil_Bit dbBits [] = {
  {"appinfoDirty", dlpDBFlagAppInfoDirty},
  {"backup",       dlpDBFlagBackup},
  {"newer",        dlpDBFlagNewer},
  {"open",         dlpDBFlagOpen},
  {"readOnly",     dlpDBFlagReadOnly},
  {"reset",        dlpDBFlagReset},
  {"resource",     dlpDBFlagResource},
  {NULL,           0}
};

static TclUtil_Bitset dbFlags = {
  "db flag",
  NULL,
  dbBits
};

/*
 *------------------------------------------------------*
 * Definition for conversion of 'CardInfo' structures
 */

static TclUtil_StructItem cardInfoItems [] = {
  ITEM       ("card",         card,         CardInfo, TUSI_Int),
  ITEM       ("version",      version,      CardInfo, TUSI_Int),
  ITEM       ("created",      creation,     CardInfo, TUSI_Int),
  ITEM       ("romsize",      romSize,      CardInfo, TUSI_Int),
  ITEM       ("ramsize",      ramSize,      CardInfo, TUSI_Int),
  ITEM       ("ramfree",      ramFree,      CardInfo, TUSI_Int),
  ITEM_CHARS ("name",         name,         CardInfo, 128),
  ITEM_CHARS ("manufacturer", manufacturer, CardInfo, 128),
  ITEM_END
};

static TclUtil_Struct structCardInfo = {
  "card",
  0, NULL,
  cardInfoItems
};

/*
 *------------------------------------------------------*
 * Definition for conversion of 'NetSyncInfo' structures
 */

static TclUtil_StructItem nsyncInfoItems [] = {
  ITEM       ("lanSync",     lanSync,        NetSyncInfo, TUSI_Int),
  ITEM_CHARS ("hostName",    hostName,       NetSyncInfo, 256),
  ITEM_CHARS ("hostAddress", hostAddress,    NetSyncInfo, 40),
  ITEM_CHARS ("hostSubnet",  hostSubnetMask, NetSyncInfo, 40),
  ITEM_END
};

static TclUtil_Struct structNetSyncInfo = {
  "netsync",
  0, NULL,
  nsyncInfoItems
};

/*
 *------------------------------------------------------*
 * Definition for conversion of 'SysInfo' structures
 */

static TclUtil_StructItem sysInfoItems [] = {
  ITEM        ("romVersion", romVersion, SysInfo, TUSI_Int),
  ITEM        ("locale",     locale,     SysInfo, TUSI_Int),
  ITEM_PCHARS ("name",       name,       SysInfo, 128, nameLength),
  ITEM_END
};

static TclUtil_Struct structSysInfo = {
  "sys",
  0, NULL,
  sysInfoItems
};

/*
 *------------------------------------------------------*
 * Definition for conversion of 'PilotUser' structures
 */

static TclUtil_StructItem puserItems [] = {
  ITEM        ("userId",             userID,             PilotUser, TUSI_Int),
  ITEM        ("viewerId",           viewerID,           PilotUser, TUSI_Int),
  ITEM        ("lastSyncPC",         lastSyncPC,         PilotUser, TUSI_Int),
  ITEM        ("successfulSyncDate", successfulSyncDate, PilotUser, TUSI_Int),
  ITEM        ("lastSyncDate",       lastSyncDate,       PilotUser, TUSI_Int),
  ITEM_CHARS  ("username",           username,           PilotUser, 128),
  ITEM_PCHARS ("password",           password,           PilotUser, 128, passwordLength),
  ITEM_END
};

static TclUtil_Struct structPilotUser = {
  "pilotuser",
  0, NULL,
  puserItems
};

/*
 *------------------------------------------------------*
 *
 *	PiTclEncodeRecAttr --
 *
 *	------------------------------------------------*
 *	Converts a list containing record attributes into
 *	the equivalent integer bitset representation.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See 'TclUtil_BitsetEncode'
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclEncodeRecAttr (interp, list, attr)
Tcl_Interp* interp;
Tcl_Obj*    list;
int*        attr;
{
  return TclUtil_BitsetEncode (interp, &recordAttributes,
			       list, attr);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclDecodeRecAttr --
 *
 *	------------------------------------------------*
 *	This procedure converts an integer containing a
 *	bitset with record attributes into a human
 *      readablee list.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See 'TclUtil_BitsetDecode'.
 *
 *	Result:
 *		A referencee to the created Tcl_Obj,
 *		or NULL
 *
 *------------------------------------------------------*
 */

Tcl_Obj*
PiTclDecodeRecAttr (interp, attr)
Tcl_Interp* interp;
int         attr;
{
  return TclUtil_BitsetDecode (interp, &recordAttributes,
			       attr);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclEncodeOpenFlags --
 *
 *	------------------------------------------------*
 *	Converts a list containing mode flags into
 *	the equivalent integer bitset representation.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See 'TclUtil_BitsetEncode'
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclEncodeOpenFlags (interp, list, flags)
Tcl_Interp* interp;
Tcl_Obj*    list;
int*        flags;
{
  return TclUtil_BitsetEncode (interp, &openFlags,
			       list, flags);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclEncodeDbFlags --
 *
 *	------------------------------------------------*
 *	Converts a list containing database attributes
 *	into the equivalent integer bitset representation.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See 'TclUtil_BitsetEncode'
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclEncodeDbFlags (interp, list, flags)
Tcl_Interp* interp;
Tcl_Obj*    list;
int*        flags;
{
  return TclUtil_BitsetEncode (interp, &dbFlags,
			       list, flags);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclDecodeDbFlags --
 *
 *	------------------------------------------------*
 *	Converts an integer containing database flags
 *	into a human readable list.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See 'TclUtil_BitsetEncode'
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

Tcl_Obj*
PiTclDecodeDbFlags (interp, flags)
Tcl_Interp* interp;
int         flags;
{
  return TclUtil_BitsetDecode (interp, &dbFlags, flags);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclCvtCardToList --
 *
 *	------------------------------------------------*
 *	This procedure converts the binary information
 *	about a memory card into a readable list. See the
 *	manpages for details on its output.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Creates tcl objects.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

Tcl_Obj*
PiTclCardToList (interp, card)
Tcl_Interp* interp;
CardInfo*   card;
{
  return TclUtil_StructDecode (interp, &structCardInfo, card);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclCvtCardsToList --
 *
 *	------------------------------------------------*
 *	This procedure converts the information about
 *	all installed cards into one big list. See the
 *	user documentation for details about the generated
 *	format.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Creates tcl objects.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

Tcl_Obj*
PiTclCardsToList (interp, dlp)
Tcl_Interp*       interp;
PiTclDlpInstance* dlp;
{
  int i;
  Tcl_Obj* list;
  Tcl_Obj* card;

  list = Tcl_NewListObj (0, NULL);

  if (list == (Tcl_Obj*) NULL) {
    return (Tcl_Obj*) NULL;
  }


  for (i=0; i < dlp->numCards; i++) {

    card = TclUtil_StructDecode (interp, &structCardInfo,
				 dlp->card [i]);

    if (card == (Tcl_Obj*) NULL) {
      Tcl_DecrRefCount (list);
      return (Tcl_Obj*) NULL;
    }

    if (Tcl_ListObjAppendElement (interp, list, card) != TCL_OK) {
      Tcl_DecrRefCount (list);
      return (Tcl_Obj*) NULL;
    }
  }

  return list;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclCvtNetSyncInfoToList --
 *
 *	------------------------------------------------*
 *	This procedure converts the netsync information of
 *	a palm pilot into a human readable string.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Creates objects.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

Tcl_Obj*
PiTclNetSyncInfoToList (interp, nsi)
Tcl_Interp*  interp;
NetSyncInfo* nsi;
{
  return TclUtil_StructDecode (interp, &structNetSyncInfo, nsi);
}


/*
 *------------------------------------------------------*
 *
 *	PiTclCvtListToNetSyncInfo --
 *
 *	------------------------------------------------*
 *	This procedure converts a human readable list back
 *	into a NetSyncInfo structure.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		May convert the incoming object into a
 *		list.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclListToNetSyncInfo (interp, list, nsi)
Tcl_Interp*  interp;
Tcl_Obj*     list;
NetSyncInfo* nsi;
{
  return TclUtil_StructEncode (interp, &structNetSyncInfo,
			       list, nsi);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclSysInfoToList --
 *
 *	------------------------------------------------*
 *	This procedure converts the system information of
 *	a palm pilot into a human readable string.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Creates objects.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

Tcl_Obj*
PiTclSysInfoToList (interp, si)
Tcl_Interp* interp;
SysInfo*    si;
{
  return TclUtil_StructDecode (interp, &structSysInfo, si);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclPilotUserToList --
 *
 *	------------------------------------------------*
 *	This procedure converts the user information of
 *	a palm pilot into a human readable string.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Creates objects.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

Tcl_Obj*
PiTclPilotUserToList  (interp, pu)
Tcl_Interp* interp;
PilotUser*  pu;
{
  return TclUtil_StructDecode (interp, &structPilotUser, pu);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclListToPilotUser --
 *
 *	------------------------------------------------*
 *	This procedure converts a human readable list back
 *	into a PilotUser structure.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		May convert the incoming object into a
 *		list.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclListToPilotUser (interp, list, pu)
Tcl_Interp*  interp;
Tcl_Obj*     list;
PilotUser*   pu;
{
  return TclUtil_StructEncode (interp, &structPilotUser,
			       list, pu);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclListDatabases --
 *
 *	------------------------------------------------*
 *	This procedure generates a list containing the
 *	names of all databases installed on the pilot the
 *	dlp object is connected to.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Creates objects, allocates memory.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclListDatabases (interp, dlp)
Tcl_Interp*       interp;
PiTclDlpInstance* dlp;
{
  Tcl_HashEntry*   hPtr;
  Tcl_HashSearch   hs;
  PiTclDbInstance* db;
  int              res;
  Tcl_Obj*         list = Tcl_NewListObj (0, NULL);

  for (hPtr  =  Tcl_FirstHashEntry (&dlp->db, &hs);
       hPtr != (Tcl_HashEntry*) NULL;
       hPtr  =  Tcl_NextHashEntry (&hs)) {

    db = ((PiTclDbInstance*) (Tcl_GetHashValue (hPtr)));

    LIST_ADDSTR (list, db->info.name);
  }

  Tcl_SetObjResult (interp, list);
  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclDbObjectToList --
 *
 *	------------------------------------------------*
 *	This procedure converts the internal database
 *	information into a human readable list. See the
 *	manpages for details on its output.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclDbObjectToList (interp, db)
Tcl_Interp*      interp;
PiTclDbInstance* db;
{
  int res;
  Tcl_Obj* list;
  Tcl_Obj* flags;
  Tcl_Obj* miscflags;

  flags     = PiTclDecodeDbFlags (interp, db->info.flags);
  miscflags = Tcl_NewListObj (0, NULL);

  if (db->info.miscFlags & dlpDBMiscFlagExcludeFromSync) {
    LIST_ADDSTR (miscflags, "excludeFromSync");
  }

  list = Tcl_NewListObj (0, NULL);

  LIST_ADDSTR (list, "name");         LIST_ADDSTR (list, db->info.name);
  LIST_ADDSTR (list, "card");         LIST_ADDINT (list, db->card);
  LIST_ADDSTR (list, "createDate");   LIST_ADDINT (list, db->info.createDate);
  LIST_ADDSTR (list, "modifyDate");   LIST_ADDINT (list, db->info.modifyDate);
  LIST_ADDSTR (list, "backupDate");   LIST_ADDINT (list, db->info.backupDate);
  LIST_ADDSTR (list, "modnum");       LIST_ADDINT (list, db->info.modnum);
  LIST_ADDSTR (list, "version");      LIST_ADDINT (list, db->info.version);
  LIST_ADDSTR (list, "type");         LIST_ADDOBJ (list, PiTclLongToString (db->info.type));
  LIST_ADDSTR (list, "creator");      LIST_ADDOBJ (list, PiTclLongToString (db->info.creator));
  LIST_ADDSTR (list, "flags");        LIST_ADDOBJ (list, flags);
  LIST_ADDSTR (list, "miscflags");    LIST_ADDOBJ (list, miscflags);

  Tcl_SetObjResult (interp, list);
  return TCL_OK;
}




