#ifndef TCL_UTIL_H
#define TCL_UTIL_H
/* -*- c -*-
 *
 * tclUtil.h --
 *
 *	General utilities helping in the construction of extensions
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: tclUtil.h,v 1.2 1999/06/28 21:20:36 aku Exp $
 */

#ifdef __cplusplus
extern "C" {
#endif

#include <string.h>
#include <tcl.h>


/*
 *------------------------------------------------------*
 * Convenient definition of conversions between bitsets
 * and lists of strings.
 */

/* The following structure describes a single position
 * in a bitset.
 */

typedef struct TclUtil_Bit {
  char* text; /* name of the bit, to be used in lists */
  int   bit;  /* value of the bit (*not* the index) */
} TclUtil_Bit;

/* The following structure describes a complete bitset.
 */

typedef struct TclUtil_Bitset {
  char*       name;  /* Name of the bitset, for use in
		      * 'Tcl_GetIndexFromObj'. */
  char**      texts; /* Reference to a dynamically allocated
		      * array referencing the names of all
		      * bits described by the structure, for
		      * use with 'Tcl_GetIndexFromObj'.
		      * Initially NULL, causing automatic
		      * creation */
  TclUtil_Bit* bits; /* Reference to a table describing all
		      * bits in the set. The table is
		      * terminated by a bit whose 'text' is
		      * NULL. */
} TclUtil_Bitset;


int
TclUtil_BitsetEncode _ANSI_ARGS_ ((Tcl_Interp*     interp,
				   TclUtil_Bitset* setDef,
				   Tcl_Obj*        setList,
				   int*            set));

Tcl_Obj*
TclUtil_BitsetDecode _ANSI_ARGS_ ((Tcl_Interp*     interp,
				   TclUtil_Bitset* setDef,
				   int             set));




/*
 *------------------------------------------------------*
 * Convenient definition of conversions between numbers
 * and strings (Enumeration)
 */

/* The following structure describes a single value
 * in an enumeration.
 */

typedef struct TclUtil_EnumItem {
  char* text;  /* name of the value, to be used in lists */
  int   value; /* enumeration value associated to the text */
} TclUtil_EnumItem;

/* The following structure describes a complete enumeration.
 */

typedef struct TclUtil_Enum {
  char*       name;        /* Name of the enumeration, for use in
			    * 'Tcl_GetIndexFromObj'. */
  char**      texts;       /* Reference to a dynamically allocated
			    * array referencing the names of all
			    * enum values described by the structure,
			    * for use with 'Tcl_GetIndexFromObj'.
			    * Initially NULL, causing automatic
			    * creation */
  TclUtil_EnumItem* value; /* Reference to a table describing all
			    * values in the enumeration. The table
			    * is terminated by a value whose 'text'
			    * is NULL. */
} TclUtil_Enum;


int
TclUtil_EnumEncode _ANSI_ARGS_ ((Tcl_Interp*   interp,
				 TclUtil_Enum* enumDef,
				 Tcl_Obj*      o,
				 int*          enValue));

Tcl_Obj*
TclUtil_EnumDecode _ANSI_ARGS_ ((Tcl_Interp*   interp,
				 TclUtil_Enum* enumDef,
				 int           enValue));



/*
 *------------------------------------------------------*
 * Convenient definition of conversion from structs into
 * key,value-lists (usable by 'array set').
 */

/* Enumeration of handled item types.
 */

typedef enum TclUtil_StructItemType {
  TUSI_Int,     /* A standard integer */
  TUSI_Boolean, /* Integer variant, contains a boolean value */
  TUSI_Double,  /* A standard double */
  TUSI_CharN,   /* inlined string, exactly 'n' characters */
  TUSI_Chars,   /* inlined string, at most 'n' characters */
  TUSI_PChars,  /* inlined string, at most 'n' characters,
		 * with the real length in a separate structure
		 * field, assumed as int. (pascal-like, hence
		 * the 'P') */
  TUSI_String,  /* A pointer to a character string */
  TUSI_Bitset,  /* A TclUtil_Bitset governs conversion, integer variant */
  TUSI_Struct,  /* Embedded substructures (*no* pointer!) */
  TUSI_Enum,    /* Integer variant, value <-> string conversion */
  TUSI_Custom,  /* Type for items currently not known,
		 * custom converter procedures are required. */
  TUSI_Byte     /* An 8-bit integer */
} TclUtil_StructItemType;

/* Interfaces used by custom conversion functions (TUSI_Custom)
 */

typedef Tcl_Obj* TclUtil_StructItemToObject
     _ANSI_ARGS_ ((Tcl_Interp* interp,
		   void*       itemPtr,
		   ClientData  clientData));

typedef int        TclUtil_StructItemFromObject
     _ANSI_ARGS_ ((Tcl_Interp* interp,
		   Tcl_Obj*    o,
		   void*       itemPtr,
		   ClientData  clientData));

/* Description of a single item in a structure.
 */

typedef struct TclUtil_StructItem {
  char* name;   /* name of the item, used as key in the lists */
  int   type;   /* item type, for selection of the appropriate
		 * converter routine. Implicitly defines the
		 * assumed size. */
  int   offset; /* offset of the item from the beginning of
		 * the structure, in bytes. */

  int   length; /* Relevant for CharN and Chars only, length of
		 * inlined string for the first, max. length of
		 * the string for the latter */

  int   lenOffset; /* offset of the 'length' item from the
		    * beginning of the structure, in bytes.
		    * Relevant to 'TUSI_PChars'. */

  /* References to custom converter functions, usually NULL.
   * Required for type 'TUSI_Custom'.
   */

  TclUtil_StructItemToObject*   toObjProc;
  TclUtil_StructItemFromObject* fromObjProc;

  ClientData clientData; /* Several meanings:
			  * 1) client data for custom converters (TUSI_Custom)
			  * 2) Reference to bitset definition    (TUSI_Bitset)
			  * 3) Reference to substructure def     (TUSI_Struct)
			  */

} TclUtil_StructItem;

/* Helpers for generation of item descriptions
 */

#ifdef offsetof
#define TU_Offset(type, field) ((int) offsetof(type, field))
#else
#define TU_Offset(type, field) ((int) ((char *) &((type *) 0)->field))
#endif

#define ITEM(name,field,struct,type) \
{(name), (type), TU_Offset(struct,field), 0, 0, NULL, NULL, NULL}

#define ITEM_CHARN(name,field,struct,n) \
{(name), TUSI_CharN, TU_Offset(struct,field), (n), 0, NULL, NULL, NULL}

#define ITEM_CHARS(name,field,struct,n) \
{(name), TUSI_Chars, TU_Offset(struct,field), (n), 0, NULL, NULL, NULL}

#define ITEM_PCHARS(name,field,struct,n,lenfield) \
{(name), TUSI_PChars, TU_Offset(struct,field), (n), TU_Offset(struct,lenfield), NULL, NULL, NULL}

#define ITEM_BITSET(name,field,struct,set) \
{(name), TUSI_Bitset, TU_Offset(struct,field), 0, 0, NULL, NULL, (ClientData) (set)}

#define ITEM_STRUCT(name,field,struct,sdef) \
{(name), TUSI_Struct, TU_Offset(struct,field), 0, 0, NULL, NULL, (ClientData) (sdef)}

#define ITEM_ENUM(name,field,struct,edef) \
{(name), TUSI_Enum, TU_Offset(struct,field), 0, 0, NULL, NULL, (ClientData) (edef)}

#define ITEM_CUSTOM(name,field,struct,to,from,cd) \
{(name), TUSI_Custom, TU_Offset(struct,field), 0, 0, (to), (from), (ClientData) (cd)}

#define ITEM_END {NULL, 0, 0, 0, 0, NULL, NULL, NULL}

/* Description of a structure.
 */

typedef struct TclUtil_Struct {
  char*               name;  /* Name of the structure, for use in
			      * 'Tcl_GetIndexFromObj'. */
  long int      allItemMask; /* Bitmask for all items, this definition
			      * restricts structures to 32, maybe 64
			      * elements ! */
  char**              texts; /* Reference to a dynamically allocated
			      * array referencing the names of all
			      * items described by the structure, for
			      * use with 'Tcl_GetIndexFromObj'.
			      * Initially NULL, causing automatic
			      * creation */
  TclUtil_StructItem* items; /* Reference to a table describing all
			      * item in the structure. The table is
			      * terminated by a item whose 'text' is
			      * NULL. */
} TclUtil_Struct;


int
TclUtil_StructEncode _ANSI_ARGS_ ((Tcl_Interp*     interp,
				   TclUtil_Struct* structDef,
				   Tcl_Obj*        structList,
				   void*           structPtr));

Tcl_Obj*
TclUtil_StructDecode _ANSI_ARGS_ ((Tcl_Interp*     interp,
				   TclUtil_Struct* structDef,
				   void*           structPtr));


/*
 *------------------------------------------------------*
 * General helper macros
 */

#define LIST_ADDOBJ(list,o) \
  res = Tcl_ListObjAppendElement (interp, list, o); \
  if (res != TCL_OK) {                              \
    Tcl_DecrRefCount (list);                        \
    return TCL_ERROR;                               \
  }

#define BIT(i) (1<<(i))
#define DUP(s) strcpy ((char*) Tcl_Alloc (strlen((s))+1), (s))


#ifdef __cplusplus
}
#endif /* C++ */
#endif /* TCL_UTIL_H */
