/*
 * piTclUtility.c --
 *
 *	Implementation of utility functions.
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: util.c,v 1.1 1999/03/27 13:26:34 aku Exp $
 */


#include "internal.h"

static int
DlpReleaseCards _ANSI_ARGS_ ((PiTclDlpInstance* dlp));

#ifndef WORDS_BIGENDIAN

#define FLIP(b,i,j) b[4] = b[i]; b[i] = b[j] ; b[j] = b[4]
#define FLIP4(b) FLIP (b,0,3); FLIP (b,1,2)

#endif


/*
 *------------------------------------------------------*
 *
 *	PiTclCreateDlpObject --
 *
 *	------------------------------------------------*
 *	This procedure creates the instance data of a new
 *	DLP object.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Allocates memory, creates a pilot socket,
 *		registers the new object in the global
 *		data structures
 *
 *	Result:
 *		A pointer to the new structure, or NULL.
 *
 *------------------------------------------------------*
 */

PiTclDlpInstance*
PiTclCreateDlpObject (info, deviceName)
PiTclInfo* info;
char*      deviceName;
{
  int                ret, sd;
  PiTclDlpInstance*  dlp;
  Tcl_HashEntry*     hPtr;

  sd = PiTclOpenDlpChannel (deviceName);

  if (sd == -1) {
    return NULL;
  }

  /* Initialize the structure as much as is possible ..
   */

  dlp = (PiTclDlpInstance*) Tcl_Alloc (sizeof (PiTclDlpInstance));

  dlp->base.info     = info;
  dlp->dev           = DUP (deviceName);
  dlp->sd            = sd;
  dlp->acceptDone    = 0;
  dlp->inSession     = 0;
  dlp->numCards      = 0;
  dlp->card          = (CardInfo**) NULL;
  dlp->dbInitialized = 0;

  Tcl_InitHashTable (&dlp->db, TCL_STRING_KEYS);

  /* .. and lock the device.
   * (ret is a dummy, we don't want it, but have to provide
   * the pointer).
   */

  hPtr = Tcl_CreateHashEntry (&info->objects, dlp->dev, &ret);
  Tcl_SetHashValue (hPtr, dlp);

  return dlp;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclOpenDlpChannel --
 *
 *	------------------------------------------------*
 *	This procedure initializes the pilot-socket for
 *	a DLP object.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Allocates memory, creates a pilot socket,
 *
 *	Result:
 *		The socket handle, or -1 in case of an error.
 *
 *------------------------------------------------------*
 */

int
PiTclOpenDlpChannel (dev)
char* dev;
{
  int                ret, sd;
  struct pi_sockaddr addr;

  /* Create a server socket, bind and listen. *Don't* 'accept'.
   * This is done later, either explicitly through the 'activate'
   * subcommand, or automatically by the first real DLP function.
   */

  sd = pi_socket (PI_AF_SLP, PI_SOCK_STREAM, PI_PF_PADP);

  if (!sd) {
    return -1;
  }

  /* -W- Uugh, the next command is prone to buffer overflow !!
   * -W- Think about checks, or automatic resize.
   */    

  strcpy (addr.pi_device, dev);
  addr.pi_family = PI_AF_SLP;
  
  ret = pi_bind (sd, (struct sockaddr*)&addr, sizeof(addr));

  if (ret == -1) {
    return -1;
  }

  ret = pi_listen (sd, 1);
  if (ret == -1) {
    return -1;
  }

  return sd;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclReleaseDlpObject --
 *
 *	------------------------------------------------*
 *	This procedure releases all memory associated to
 *	a dlp instance structure, as allocated by
 *      'PiTclCreateDlpObject' and others.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Releases memory, does not touch the
 *		socket, expects that this one is closed
 *		before calling this procedure.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclReleaseDlpObject (dlp)
PiTclDlpInstance* dlp;
{
  Tcl_HashEntry*    hPtr;

  /* Remove the object from the global hash table, release the device.
   */

  if (dlp->base.info != (PiTclInfo*) NULL) {
    hPtr = Tcl_FindHashEntry (&dlp->base.info->objects, dlp->dev);

    if (hPtr != (Tcl_HashEntry*) NULL) {
      Tcl_DeleteHashEntry (hPtr);
    }
  }

  /* Release memory allocated during the lifetime of the object.
   */

  DlpReleaseCards (dlp);


  /* Release the local cahce of database information.
   */

  Tcl_DeleteHashTable (&dlp->db);

  /* Deregister the reference to the main datastructure.
   */

  PiTclDeleteCommand (&dlp->base.info);

  /* And the main object information.
   */

  Tcl_Free (dlp->dev);
  Tcl_Free ((char*) dlp);

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclDlpReadCards --
 *
 *	------------------------------------------------*
 *	This procedure uses the DLP layer to retrieve
 *	everything the information about all memory cards
 *	installed in the pilot we are connected to.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Allocates memory.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclDlpReadCards (dlp, cmd, interp)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
{
  CardInfo c;
  int      dlpResult, i;

  if (dlp->card != (CardInfo**) NULL) {
    /* Cards already read from the pilot.
     */
    return TCL_OK;
  }

  /* Iterate over all cards in the pilot and store their
   * description in the instance record.
   */

  for (i=0;;i++) {
    dlpResult = dlp_ReadStorageInfo (dlp->sd, i, &c);

    if (dlpResult == dlpErrNotFound) {
      /* No more cards */
      return TCL_OK;
    }

    if (dlpResult < 0) {
      Tcl_AppendResult (interp,
			cmd,
			" error while reading card information:",
			dlp_strerror (dlpResult), NULL);
      return TCL_ERROR;     
    }

    dlp->numCards ++;

    if (dlp->card ==  (CardInfo**) NULL) {
      dlp->card = (CardInfo**) Tcl_Alloc (sizeof (CardInfo*));
    } else {
      int size = dlp->numCards * sizeof (CardInfo*);

      dlp->card = (CardInfo**) Tcl_Realloc ((char*) dlp->card, size);
    }

    dlp->card [i] = (CardInfo*) Tcl_Alloc (sizeof (CardInfo));
    memcpy (dlp->card [i], &c, sizeof (CardInfo));

    if (! c.more) {
      /* This is not reliable, checking for 'dlpErrNotFound' too.
       */
      break;
    }
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DlpReleaseCards --
 *
 *	------------------------------------------------*
 *	This procedure releases all memory associated to
 *	memory cards in a dlp object.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Releases the memory allocated in
 *		'PiTclDlpReadCards'.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DlpReleaseCards (dlp)
PiTclDlpInstance* dlp;
{
  if (dlp->card != (CardInfo**) NULL) {
    int i;

    for (i=0; i < dlp->numCards; i++) {
      Tcl_Free ((char*) dlp->card [i]);
    }

    Tcl_Free ((char*) dlp->card);
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclDlpReadDbList --
 *
 *	------------------------------------------------*
 *	This procedure talks to the pilot and retrieves
 *	a list of all databases it knows about. This is
 *	then used to built a local cache of this infor-
 *	mation for fast access by the database operations.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Allocates memory.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclDlpReadDbList (dlp, cmd, interp)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
{
  DBInfo dbInfo;
  int card, dlpResult, idx;


  if (dlp->dbInitialized) {
    /* The cache is already built, no need to do it again. */
    return TCL_OK;
  }

#define FLAGS (dlpDBListRAM | dlpDBListROM)

  /* Iterate over all memory cards and all the databases
   * in them, but only after we know everything about the
   * installed cards.
   */

  if (PiTclDlpReadCards (dlp, cmd, interp) != TCL_OK) {
    return TCL_ERROR;
  }

  for (card = 0; card < dlp->numCards; card++) {

    idx = 0;
    while (1) {
      dlpResult = dlp_ReadDBList (dlp->sd, card, FLAGS, idx, &dbInfo);
      /*                        Get the entry at index 'idx' */

      if (dlpResult == dlpErrNotFound) {
	/* No more databases on this card */
	break;
      }

      if (dlpResult < 0) {
	/* Construct error message from the return value of
	 * the DLP function.
	 */

	Tcl_AppendResult (interp,
			  cmd," db (reading databases):",
			  dlp_strerror (dlpResult),
			  (char*) NULL);

	return TCL_ERROR;
      }

      /* Add the now known database to our local cache.
       * (Discard the function result, not required here)
       */

      PiTclCreateDbObject (dlp, &dbInfo, card);

      /* Prepare for next round
       */

      idx = dbInfo.index + 1;
    }
  }

#undef FLAGS

  dlp->dbInitialized = 1;
  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclCreateDbObject --
 *
 *	------------------------------------------------*
 *	This procedure allocates and initializes a
 *	database onject structure.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Allocates memory.
 *
 *	Result:
 *		A reference to the new object, or NULL.
 *
 *------------------------------------------------------*
 */

PiTclDbInstance*
PiTclCreateDbObject (dlp, dbInfo, card)
PiTclDlpInstance* dlp;
DBInfo*           dbInfo;
int               card;
{
  PiTclDbInstance* db;
  int             ret;
  Tcl_HashEntry*  hPtr;

  db = (PiTclDbInstance*) Tcl_Alloc (sizeof (PiTclDbInstance));

  db->dlp      = dlp;
  db->sd       = -1;                 /* Not open! */
  db->cmd      = (Tcl_Command) NULL; /* Not open! */
  db->handle   = -1;                 /* Not open! */
  db->card     = card;
  db->pack     = (PiTclPackInfo*) NULL;
  db->buf      = Tcl_Alloc (MAXBUF);

  memcpy (&db->info, dbInfo, sizeof (DBInfo));

  db->name     = Tcl_NewStringObj  (db->info.name, -1);
  db->creator  = PiTclLongToString (db->info.creator);
  db->type     = PiTclLongToString (db->info.type);

  Tcl_IncrRefCount (db->name);
  Tcl_IncrRefCount (db->creator);
  Tcl_IncrRefCount (db->type);

  hPtr = Tcl_CreateHashEntry (&dlp->db, dbInfo->name, &ret);
  Tcl_SetHashValue (hPtr, db);

  return db;
}

/*
 *------------------------------------------------------*
 *
 *	DlpDbRelease --
 *
 *	------------------------------------------------*
 *	This procedure releases any information allocated
 *	during the construction of an entry of the local
 *	database cache.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Releases memory.
 *
 *	Result:
 *		Nothing.
 *
 *------------------------------------------------------*
 */

void
PiTclReleaseDbObject (db)
PiTclDbInstance* db;
{
  Tcl_DecrRefCount (db->name);
  Tcl_DecrRefCount (db->creator);
  Tcl_DecrRefCount (db->type);

  if (db->pack != (PiTclPackInfo*) NULL) {
    PiTclPackRelease (db->pack);
  }

  Tcl_Free (db->buf);
  Tcl_Free ((char*) db);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclDlpDbCloseAll --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'db closeall' sub-
 *	command of dlp objects. See the manpages for
 *	details on what it does. It additionally allows
 *	the destruction of the local cache of db objects.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclDlpDbCloseAll (dlp, cmd, interp, mode)
PiTclDlpInstance* dlp;
char*             cmd;
Tcl_Interp*       interp;
int               mode;
{
  /* Two modes possible:
   * PITCL_DB_CLOSE:   Closes all open databases.
   * PITCL_DB_DESTROY: As above, additionally destroys
   *                   the local cache of database objects.
   */

  Tcl_HashSearch   hs;
  Tcl_HashEntry*   h;
  /*  int              dlpResult; see #if 0 below */
  PiTclDbInstance* db;

  if (!dlp->dbInitialized) {
    /* We have no local cache, i.e. we never talked db
     * operations to the pilot! Nothing has to be done,
     * for neither mode.
     */

    return TCL_OK;
  }

  /* An initialized database cache is possible only for
   * an active connection to the pilot. No need to check
   * that.
   * --     CHK_ACTIVE(dlp, cmd, interp);
   */
#if 0
  /* PROBLEMS, error despite valid socket */
  dlpResult = dlp_CloseDB_All (dlp->sd);

  if (cmd != (char*) NULL) {
    /* A command reference containing NULL indicates a
     * call during the destruction of the dlp object.
     * A possible error is not returned then.
     */

    DLP_CHECK (interp, cmd, " db closeAll ", dlpResult);
  }
#endif
  /* Get the database objects and break their connection to the
   * pilot. The destruction of the object command will then
   * refrain from talking to the pilot. If the caller requested
   * the additional destruction of all db structures this is
   * then done too.
   */

  for (h = Tcl_FirstHashEntry (&dlp->db, &hs);
       h != (Tcl_HashEntry*) NULL;
       h = Tcl_NextHashEntry (&hs)) {

    db = ((PiTclDbInstance*) (Tcl_GetHashValue (h)));

    if (db->cmd != (Tcl_Command) NULL) {
      /* Database is open, close it, only on the client side.
       * The pilot has already closed it, due to the call to
       * 'dlp_CloseDB_All'. The assignment below enforces that
       * the object cleanup does not talk to the pilot.
       */

      /* db->sd = -1; */ /* Close! db here !, not via 'all' */

      Tcl_DeleteCommandFromToken (interp, db->cmd);
    }

    if (mode == PITCL_DB_DESTROY) {
      /* The caller requested destruction of the local cache,
       * do that now too.
       */

      Tcl_DeleteHashEntry  (h);
      PiTclReleaseDbObject (db);
    }
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclStringToLong --
 *
 *	------------------------------------------------*
 *	This procedure converts a string into a long
 *	value, truncating or padding with '\0', as is
 *	necessary. Usages: Resource Type, Database
 *	Creator and Type.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		None.
 *
 *	Result:
 *		A long value.
 *
 *------------------------------------------------------*
 */

long int
PiTclStringToLong (o)
Tcl_Obj* o;
{
  char     tmp [5];
  long int res;
  int      len;
  char*    s = Tcl_GetStringFromObj (o, &len);

  /* The string is always bigendian. Flip the information
   * if we are on a little-endian machine to get the
   * correct integer representation.
   */

  if (len < 4) {
    memset (&tmp, '\0', 4);
    memcpy (&tmp, s, len);
  } else {
    memcpy (&tmp, s, 4);
  }

#ifndef WORDS_BIGENDIAN
  FLIP4 (tmp);
#endif

  memcpy (&res, tmp, 4);

  return res;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclLongToString --
 *
 *	------------------------------------------------*
 *	This procedure converts a long value into a string
 *	Usages: Resource Type, Database Creator and Type.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Creates an Tcl_Obj'ect, allocates memory
 *
 *	Result:
 *		A Tcl_Obj reference.
 *
 *------------------------------------------------------*
 */

Tcl_Obj*
PiTclLongToString (l)
long int l;
{
#ifndef WORDS_BIGENDIAN
    char buf [5];

    memcpy (buf, (char*) &l, 4);
    FLIP4 (buf);

    return Tcl_NewStringObj (buf, 4);
#else
    return Tcl_NewStringObj ((char*) &l, 4);
#endif
}

